package edu.uiuc.ncsa.security.servlet.mail;

import edu.uiuc.ncsa.security.core.Logable;
import edu.uiuc.ncsa.security.core.exceptions.GeneralException;
import edu.uiuc.ncsa.security.core.util.AbstractEnvironment;
import edu.uiuc.ncsa.security.core.util.MyLoggingFacade;

import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import javax.naming.Context;
import javax.naming.InitialContext;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.Map;
import java.util.Properties;

/**
 * A utility for sending messages via SMTP or SMTPS using Java mail.
 * <p>Created by Jeff Gaynor<br>
 * on 10/5/11 at  1:18 PM
 */
public class MailUtil implements Logable {
    public MailUtil(MyLoggingFacade myLogger) {
        this.myLogger = myLogger;
    }

    /**
     * The left-hand delimiter for replacements in templates. Must make a valid regex when combined with the
     * key and right-hand delimiter.
     */
    public static String LEFT_DELIMITER = "\\$\\{";
    /**
     * The right-hand delimiter for replacements in templates. Must make a valid regex when combined with the
     * left-hand delimiter and key.
     */
    public static String RIGHT_DELIMITER = "\\}";
    /**
     * The separator between email addresses. The default is a semi-colon.
     */
    public static String ADDRESS_SEPARATOR = ";";


    public MailUtil() {
        try {
            mailEnvironment = new MailEnvironment(false, null, -1, null, null, null, null, null, false, false);
        } catch (Exception x) {
            warn("Error: could not create mail environment:" + x);
        }
    }

    public MailUtil(MailEnvironment me) {
        mailEnvironment = me;
    }

    public boolean isEnabled() {
        return getMailEnvironment().mailEnabled;
    }

    public static class MailEnvironment extends AbstractEnvironment {
        public MailEnvironment(boolean mailEnabled) {
            this.mailEnabled = mailEnabled;
        }

        boolean starttls = false;

        public MailEnvironment(
                boolean mailEnabled,
                String server,
                int port,
                String password,
                String from,
                String recipients,
                String messageTemplate,
                String subjectTemplate,
                boolean useSSL,
                boolean starttls) throws Exception {
            if (from != null) {
                this.from = new InternetAddress(from);
            }
            this.mailEnabled = mailEnabled;
            this.messageTemplate = messageTemplate;
            this.password = password;
            this.port = port;
            this.recipients = parseRecipients(recipients);
            this.server = server;
            this.subjectTemplate = subjectTemplate;
            this.useSSL = useSSL;
            this.starttls = starttls;
        }

        protected void init(boolean mailEnabled,
                            String server,
                            int port,
                            String password,
                            String from,
                            String recipients,
                            String messageTemplate,
                            String subjectTemplate,
                            boolean useSSL,
                            boolean starttls) throws AddressException {
            this.mailEnabled = mailEnabled;
            info("Mail notifications " + (mailEnabled ? "en" : "dis") + "abled");
            this.useSSL = useSSL;
            this.starttls = starttls;
            if (from != null) {
                this.from = new InternetAddress(from);
            }

            if (subjectTemplate == null) {
                warn("Email: No subject template found, using default");
                subjectTemplate = "Something happened on the server.";
            } else {
                this.subjectTemplate = subjectTemplate;
            }

            this.server = server;
            if (messageTemplate == null) {
                warn("Email: No message template found, using default");
                messageTemplate = "Something happened on the server.";
            } else {
                this.messageTemplate = messageTemplate;
            }

            this.port = port;
            this.password = password;
            this.recipients = parseRecipients(recipients);
            debug("Email uses ssl: " + (useSSL ? "y" : "n") + ", server=" + server + ", sender=" + from);

        }


        boolean mailEnabled;
        boolean useSSL;


        InternetAddress from = null;
        String password;
        String server;
        int port = -1;
        String messageTemplate;
        String subjectTemplate;
        public InternetAddress[] recipients;

        protected InternetAddress[] parseRecipients(String x) throws AddressException {
            InternetAddress[] z = null;
            if (x != null) {
                String[] addresses = x.split(ADDRESS_SEPARATOR);
                z = new InternetAddress[addresses.length];
                int i = 0;
                for (String y : addresses) {
                    z[i++] = new InternetAddress(y);
                }
            }
            if (z == null) {
                warn("Warning: No recipients found for email notification");
                z = new InternetAddress[0];
            }
            return z;
        }
    }

    public MailEnvironment getMailEnvironment() {
        return mailEnvironment;
    }

    MailEnvironment mailEnvironment;

    public static final String DOMAIN = "edu.uiuc.ncsa.security.servlet.mail";
    /**
     * Set this to true (case insensitive) to enable this feature. Any other value disables it.
     */
    public static final String EMAIL_ENABLED = DOMAIN + ".enabled";
    /**
     * Set to true (case insensitive) to use SSL. Any other value uses standard SMTP.
     */
    public static final String EMAIL_SSL = DOMAIN + ".ssl";
    /**
     * The user name to user when connecting to the mail server
     */
    public static final String EMAIL_USER_NAME = DOMAIN + ".user";
    /**
     * The user's password
     */
    public static final String EMAIL_PASSWORD = DOMAIN + ".password";
    /**
     * The mail server.
     */
    public static final String EMAIL_SERVER = DOMAIN + ".server";
    /**
     * The port to use. Default is 25 for SMTP and 465 for SMTPS. If you do not supply
     * a port, the default for that protocol will be used.
     */
    public static final String EMAIL_PORT = DOMAIN + ".port";
    /**
     * List of recipients, separated with semi-colons.
     */
    public static final String EMAIL_RECIPIENTS = DOMAIN + ".recipients";
    /**
     * File name of template to use for messages
     */
    public static final String EMAIL_MESSAGE_TEMPLATE = DOMAIN + ".message";
    /**
     * File name of template to use for the subject. Omitting this uses a generic subject.
     */
    public static final String EMAIL_SUBJECT_TEMPLATE = DOMAIN + ".subject";

    /**
     * This takes a map of replacements for the templates and sends the message.
     * <p>This will return a true if the message succeeded and a false otherwise.
     * It will <b>not</b> cause a failure outright, since a failed notification should not
     * bring down your server.</p>
     *
     * @param replacements
     * @return
     */
    public boolean sendMessage(Map replacements) {
        Transport tr = null;

        try {
            Context initCtx = new InitialContext();
            Context envCtx = (Context) initCtx.lookup("java:comp/env");
            Object obj = envCtx.lookup("mail/Session");

            /*
           If you get an exception here like
           ClassCastException: cannot cast javax.mail.Session to javax.mail.Session
           i.e., same class to class, this is because they came from different class loaders.
           Tomcat should have a copy of whatever java mail jar in its $CATALINA_HOME/lib. This is
           available at runtime to this servlet. If you add this jar to the webapp and deploy
           is in the webapp's lib directory, you will get this Exception.

           Solution is to have only the single copy of the java-mail jar in the $CATALINA_HOME/lib directory.
           You will need to include in your compile classpath when compiling though. This is done in maven
           by specifying the java mail jar as a dependency with scope "provided"
            */
            Session session = (Session) obj;

            info("Preparing to send mail notification");
            Message message = new MimeMessage(session);
            message.setFrom(getMailEnvironment().from);
            message.setRecipients(Message.RecipientType.TO, getMailEnvironment().recipients);
            message.setSubject(replaceAll2(getSubjectTemplate(), replacements));
            message.setContent(replaceAll2(getMessageTemplate(), replacements), "text/plain");

            Properties props = new Properties();
            if (getMailEnvironment().isDebugOn()) {
                props.put("mail.debug", "true");
            }
            String protocol;
            int defaultPort = getMailEnvironment().port;
            if (getMailEnvironment().useSSL) {
                protocol = "smtps";
                if (getMailEnvironment().starttls) {
                    props.put("mail.smtp.starttls.enable", "true"); // fix for gmail, mostly. .
                }
                defaultPort = defaultPort == -1 ? 465 : defaultPort;
            } else {
                protocol = "smtp";
                defaultPort = defaultPort == -1 ? 25 : defaultPort;
            }
            debug("preparing message with protocol=" + protocol + " on server=" + getMailEnvironment().server);
            props.put("mail.transport.protocol", protocol);
            props.put("mail." + protocol + ".host", getMailEnvironment().server);
            if (protocol.equals("smtp") && getMailEnvironment().from == null) {
                props.put("mail." + protocol + ".auth", "false");

            } else {
                debug("Using authorization for user " + getMailEnvironment().from);
                props.put("mail." + protocol + ".auth", "true");
            }
            session = Session.getDefaultInstance(props);
            tr = session.getTransport(protocol);
            tr.connect(getMailEnvironment().server,
                    defaultPort,
                    getMailEnvironment().from == null ? null : getMailEnvironment().from.toString(),
                    getMailEnvironment().password);
            tr.sendMessage(message, getMailEnvironment().recipients);
            info("Mail notification sent to " + Arrays.toString(getMailEnvironment().recipients));
            return true;
        } catch (Throwable throwable) {
            info("got exception sending message:");
            for(Object key : replacements.keySet()){
                info("(" + key + "," + replacements.get(key.toString()) + ")");
            }
            throwable.printStackTrace();
            error("Sending mail failed. Continuing & message reads \"" + throwable.getMessage() + "\"");
            return false;
            //throw new GeneralException("Error", throwable);
        } finally {
            if (tr != null) {
                try {
                    tr.close();
                } catch (MessagingException e) {
                    throw new GeneralException("Error: could not close java mail transport", e);
                }
            }
        }
    }

    public String getMessageTemplate() throws IOException {
        if (messageTemplate == null) {
            messageTemplate = readTemplate(getMailEnvironment().messageTemplate);
        }
        return messageTemplate;
    }

    public void setMessageTemplate(String messageTemplate) {
        this.messageTemplate = messageTemplate;
    }

    public void setSubjectTemplate(String subjectTemplate) {
        this.subjectTemplate = subjectTemplate;
    }

    String messageTemplate;

    public String getSubjectTemplate() throws IOException {
        if (subjectTemplate == null) {
            subjectTemplate = readTemplate(getMailEnvironment().subjectTemplate);
        }
        return subjectTemplate;
    }

    String subjectTemplate;

    /**
     * Simple-minded template replacement. This works well for small, simple  arguments. To scale this up, use reg-exs and
     * be prepared for some tricky edge cases.
     * @param template
     * @param replacements
     * @return
     */
    protected String replaceAll2(String template, Map replacements) {
        String out = template;
        int count = 0;

        for (Object key : replacements.keySet()) {
            // Have to properly escape the regex here.
            count++;
            String newKey = "${" + key.toString() + "}";
            int indexOf = out.indexOf(newKey);
            while (0 <= indexOf) {
                String tail = "";
                if(out.length() > indexOf + newKey.length() + 1){
                    tail = out.substring(indexOf + newKey.length());
                }
                out = out.substring(0, indexOf) + replacements.get(key).toString() + tail ;
                indexOf = out.indexOf(newKey);
            }
        }
        debug("made " + count + " replacements in the template");
        return out;
    }

    protected String readTemplate(String fileName) throws IOException {
        debug("reading in template file \"" + fileName + "\"");
        BufferedReader br = new BufferedReader(new FileReader(fileName));
        String out = null;
        String inLine = br.readLine();
        boolean isFirst = true;
        while (inLine != null) {
            if (!inLine.startsWith("#")) {
                if (isFirst) {
                    isFirst = false;
                    out = inLine;
                } else {
                    out = out + "\n" + inLine;
                }
            }
            inLine = br.readLine();
        }
        br.close();
        return out;
    }

    public MyLoggingFacade getMyLogger() {
        if (myLogger == null) {
            myLogger = new MyLoggingFacade(getClass().getName());
        }
        return myLogger;
    }

    MyLoggingFacade myLogger;

    public void debug(String x) {
        getMyLogger().debug(x);
    }

    public void setDebugOn(boolean debugOn) {
        getMyLogger().setDebugOn(debugOn);
    }


    public boolean isDebugOn() {
        return getMyLogger().isDebugOn();
    }


    public void info(String x) {
        getMyLogger().info(x);

    }

    public void warn(String x) {
        getMyLogger().warn(x);
    }

    public void error(String x) {
        getMyLogger().error(x);
    }
}
